"""Init data"""
from __future__ import annotations

from importlib.metadata import version

from raspa_ase.calculator import Raspa

__all__ = ["Raspa"]

# Load the version
__version__ = version("raspa_ase")



from __future__ import annotations

import re
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING

from ase.atoms import Atoms
from pymatgen.io.ase import AseAtomsAdaptor

from raspa_ase.utils.dicts import convert_booleans

if TYPE_CHECKING:
    from typing import Any


def write_simulation_input(
    parameters: dict[str, Any], input_filepath: str | Path
) -> str:
    """
    Write the `simulation.input` file for a given set of parameters.

    Parameters
    ----------
    parameters
        The parameters to write to the simulation input file.
    input_filepath
        The path to the simulation input file.

    Returns
    -------
    str
        The simulation input string.
    """

    simulation_input = ""
    parameters = convert_booleans(parameters)

    for k, v in parameters.items():
        if isinstance(v, dict):
            simulation_input += f"{k}\n"
            simulation_input += _dict_to_str(v)
        elif isinstance(v, list):
            simulation_input += f"{k} "
            simulation_input += _iterable_to_str(v)
        else:
            simulation_input += f"{k} {v}\n"

    with Path(input_filepath).open(mode="w") as fd:
        fd.write(simulation_input)


def write_frameworks(frameworks: list[Atoms], directory: str | Path) -> None:
    """
    Write the CIF files for a list of frameworks.

    Parameters
    ----------
    frameworks
        The frameworks to write.
    directory
        The directory to write the CIF files to.

    Returns
    -------
    None
    """
    for i, framework in enumerate(frameworks):
        if framework == Atoms():
            continue
        name = f"framework{i}"

        structure = AseAtomsAdaptor.get_structure(framework)
        structure.to(str(Path(directory, name + ".cif")), write_site_properties=True)


def parse_output(filepath: str | Path) -> dict[str, Any]:
    """
    Specific parsing of the output file. Adapted from the following:
    https://github.com/iRASPA/RASPA2/blob/master/python/output_parser.py

    Parameters
    ----------
    filepath
        The path to the RASPA output file.

    Returns
    -------
    dict
        The parsed output data.
    """

    def _clean(split_list: list[str]) -> list[float]:
        """Strips and attempts to convert a list of strings to floats."""

        def try_float(s):
            try:
                return float(s)
            except ValueError:
                return s

        return [try_float(s.strip()) for s in split_list if s]

    with Path(filepath).open(mode="r") as fd:
        raspa_output = fd.read()

    # Reads the string into a newline-separated list, skipping useless lines
    data = [
        row.strip()
        for row in raspa_output.splitlines()
        if row and all(d not in row for d in ["-----", "+++++"])
    ]

    # Generally, categories in the output are delimited by equal signs
    delimiters = [
        i
        for i, row in enumerate(data)
        if "=====" in row and "Exclusion constraints energy" not in data[i - 1]
    ]

    # Append a row for "absolute adsorption:" and "excess adsorption:"
    # These values are separated into two rows
    abs_adsorp_rows = [i for i, row in enumerate(data) if "absolute adsorption:" in row]
    for row in abs_adsorp_rows:
        data[row] += "  " + data[row + 1]
        data[row + 2] += data[row + 3]
        data[row + 1], data[row + 3] = " ", " "

    # Use the delimiters to make a high-level dict. Title is row before
    # delimiter, and content is every row after delimiter, up to the next title
    info = {
        data[n - 1].strip(":"): data[n + 1 : delimiters[i + 1] - 1]
        for i, n in enumerate(delimiters[:-1])
    }

    # Let's PARSE!
    for key, values in info.items():
        d, note_index = {}, 1
        for item in values:
            # Takes care of all "Blocks[ #]", skipping hard-to-parse parts
            if (
                "Block" in item
                and "Box-lengths" not in key
                and "Van der Waals:" not in item
            ):
                blocks = _clean(item.split())
                d["".join(blocks[:2])] = blocks[2:]

            # Most of the average data values are parsed in this section
            elif (
                any(s in item for s in ["Average     ", "Surface area:"])
                and "desorption" not in key
            ):
                average_data = _clean(item.split())
                # Average values organized by its unit, many patterns here
                if len(average_data) == 8:
                    del average_data[2:4]
                    d[" ".join(average_data[4:6])] = average_data[1:4]
                elif len(average_data) == 5:
                    d[average_data[-1]] = average_data[1:4]
                elif "Surface" in average_data[0]:
                    d[average_data[-1]] = average_data[2:5]
                # This is the common case
                else:
                    del average_data[2]
                    d[average_data[-1]] = average_data[1:4]

            # Average box-lengths has its own pattern
            elif "Box-lengths" in key:
                box_lengths = _clean(item.split())
                i = 3 if "angle" in item else 2
                d[" ".join(box_lengths[:i])] = box_lengths[i:]

            # "Heat of Desorption" section
            elif "desorption" in key:
                if "Note" in item:
                    notes = re.split(r"[:\s]{2,}", item)
                    d["%s %d" % (notes[0], note_index)] = notes[1]
                    note_index += 1
                else:
                    heat_desorp = _clean(item.split())
                    # One line has "Average" in front, force it to be normal
                    if "Average" in item:
                        del heat_desorp[0]
                    d[heat_desorp[-1]] = heat_desorp[0:3]

            # Parts where Van der Waals are included
            elif (
                "Host-" in key or "-Cation" in key or "Adsorbate-Adsorbate" in key
            ) and "desorption" not in key:
                van_der = item.split()
                # First Column
                if "Block" in van_der[0]:
                    sub_data = [
                        _clean(s.split(":")) for s in re.split(r"\s{2,}", item)[1:]
                    ]
                    sub_dict = {s[0]: s[1] for s in sub_data[:2]}
                    d["".join(van_der[:2])] = [float(van_der[2]), sub_dict]
                # Average for each columns
                elif "Average" in item:
                    avg = _clean(re.split(r"\s{2,}", item))
                    vdw, coulomb = (_clean(s.split(": ")) for s in avg[2:4])
                    d[avg[0]] = avg[1]
                    d["Average %s" % vdw[0]] = vdw[1]
                    d["Average %s" % coulomb[0]] = coulomb[1]
                else:
                    d["standard deviation"] = _clean(van_der)

            # IMPORTANT STUFF
            elif "Number of molecules" in key:
                adsorb_data = _clean(item.rsplit(" ", 12))
                if "Component" in item:
                    gas_name = adsorb_data[2].strip("[]")
                    d[gas_name] = {}
                else:
                    d[gas_name][adsorb_data[0]] = adsorb_data[1:]

            # Henry and Widom
            elif "Average Widom" in item:
                d["Widom"] = _clean(item.rsplit(" ", 5))[1:]

            elif "Average Henry" in item:
                d["Henry"] = _clean(item.rsplit(" ", 5))[1:]

            # Ignore these
            elif any(
                s in item
                for s in ["=====", "Starting simulation", "Finishing simulation"]
            ):
                continue

            # Other strings
            else:
                parsed_data = _clean(re.split(r"[()[\]:,\t]", item))
                d[parsed_data[0]] = parsed_data[1:]
        # Putting subdictionary back into main object
        info[key] = d

    return info


def _iterable_to_str(v: list[Any]) -> str:
    """
    Convert a list to a space-separated string.

    Parameters
    ----------
    v
        The list to convert.

    Returns
    -------
    str
        The space-separated string.
    """
    return " ".join([str(i) for i in v]) + "\n"


def _dict_to_str(d: dict[str, Any]) -> str:
    """
    Convert a dictionary to a formatted string.

    Parameters
    ----------
    d
        The dictionary to convert.

    Returns
    -------
    str
        The formatted string.
    """
    s = ""
    for k, v in d.items():
        if isinstance(v, dict):
            s += f"    {k}\n    " + _dict_to_str(v)
        elif isinstance(v, Iterable) and not isinstance(v, str):
            s += f"    {k} " + _iterable_to_str(v)
        else:
            s += f"    {k} {v}\n"
    return s



"""
ASE calculator for RASPA
"""
from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING

from ase.calculators.genericfileio import (
    BaseProfile,
    CalculatorTemplate,
    GenericFileIOCalculator,
)

from raspa_ase.utils.dicts import merge_parameters, pop_parameter
from raspa_ase.utils.io import parse_output, write_frameworks, write_simulation_input
from raspa_ase.utils.params import get_framework_params

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms

SIMULATION_INPUT = "simulation.input"
LABEL = "raspa"


class RaspaProfile(BaseProfile):
    """
    RASPA profile, which defines the command that will be executed and where.
    """

    def __init__(self, binary: Path | str | None = None, **kwargs) -> None:
        """
        Initialize the RASPA profile. $RASPA_DIR must be set in the environment.

        Parameters
        ----------
        binary
            The binary to run RASPA. This defaults to doing `${RASPA_DIR}/bin/simulate`
            and typically does not need to be changed.

        Returns
        -------
        None
        """
        super().__init__(**kwargs)
        if not binary:
            raspa_dir = os.environ.get("RASPA_DIR")
            if not raspa_dir:
                raise OSError("RASPA_DIR environment variable not set")
            binary = f"{raspa_dir}/bin/simulate"
        self.binary = binary

    def get_calculator_command(self, inputfile: str = SIMULATION_INPUT) -> list[str]:
        """
        Construct the command for the calculator.

        Parameters
        ----------
        inputfile
            The name of the input file to use.

        Returns
        -------
        list[str]
            The command to run the calculator.
        """
        return [self.binary, f"{inputfile}"]

    def version(self) -> str:
        """
        Get the RASPA version number.

        Returns
        ------
        str
            The RASPA version.
        """
        raise NotImplementedError


class RaspaTemplate(CalculatorTemplate):
    """
    RASPA template, used to define how to read and write RASPA files.
    """

    def __init__(self, frameworks: list[Atoms] | None = None) -> None:
        """
        Initialize the RASPA template.

        Parameters
        ----------
        frameworks
            The frameworks to use, to be used in place of the `Atoms` object
            the calculator is applied to.

        Returns
        -------
        None
        """
        super().__init__(
            name="raspa",
            implemented_properties=["energy"],
        )

        self.inputname = SIMULATION_INPUT
        self.outputname = "raspa.stdout"
        self.errorfile = "raspa.stderr"
        self.frameworks = frameworks

    def write_input(
        self,
        profile: RaspaProfile,  # skipcq: PYL-W0613
        directory: Path | str,
        atoms: Atoms,
        parameters: dict[str, Any],
        properties: Any,  # skipcq: PYL-W0613
    ) -> None:
        """
        Write the RASPA input files.

        Parameters
        ----------
        directory
            The path to the directory to write the RASPA input files in.
        atoms
            The ASE atoms object to use as the framework.
        parameters
            The RASPA parameters to use, formatted as a dictionary.
        profile
             This is needed the base class and should not be explicitly specified.
        properties
            This is needed the base class and should not be explicitly specified.

        Returns
        -------
        None
        """
        frameworks = self.frameworks if self.frameworks else [atoms]
        parameters = merge_parameters(parameters, get_framework_params([atoms]))

        write_simulation_input(parameters, directory / self.inputname)
        write_frameworks(frameworks, directory)

    def execute(self, directory: Path | str, profile: RaspaProfile) -> None:
        profile.run(
            directory,
            self.inputname,
            directory / self.outputname,
            errorfile=directory / self.errorfile,
        )

    @staticmethod
    def read_results(directory: Path | str) -> dict[str, Any]:
        """
        Read the results of a RASPA calculation.

        Parameters
        ----------
        directory
            The path to the directory to read the RASPA results from.

        Returns
        -------
        Results
            The RASPA results, formatted as a dictionary.
        """
        output_path = Path(directory) / "Output"
        systems = Path(output_path).glob("System_*")
        results = {"energy": None}
        for system in systems:
            data_files = Path(system).glob("*.data")
            results[system.name] = {}
            for data_file in data_files:
                output = parse_output(data_file)
                results[system.name][data_file.name] = output
        return results

    def load_profile(self, cfg, **kwargs) -> RaspaProfile:
        """
        Load the RASPA profile.

        Parameters
        ----------
        cfg
            The ASE configuration profile, if any.
        **kwargs
            Any additional arguments to pass to the RASPA profile.

        Returns
        -------
        RaspaProfile
            The RASPA profile.
        """
        raise NotImplementedError


class Raspa(GenericFileIOCalculator):
    """
    The RASPA calculator.
    """

    def __init__(
        self,
        profile: RaspaProfile | None = None,
        directory: Path | str = ".",
        boxes: list[dict[str, Any]] | None = None,
        components: list[dict[str, Any]] | None = None,
        parameters: dict[str, Any] | None = None,
        multiple_frameworks: list[Atoms] | None = None,
        **kwargs,
    ) -> None:
        """
        Initialize the RASPA calculator. The $RASPA_DIR environment variable must
        be set.

        This calculator is to be set on an `Atoms` object, which will be the
        framework. All framework-related parameters should be set in the `Atoms.info`
        dictionary.

        If you have an ASE `Atoms` object of the MFI zeolite ("atoms"), then:

        ```python
        atoms.info = {"HeliumVoidFraction": 0.29}
        ```

        would be written out as the following:

        ```
        Framework 0
            FrameworkName framework0
            UnitCells 12 12 12
            HeliumVoidFraction 0.29
        ```

        The `FrameworkName` is automaticallly set by the calculator.
        The UnitCells parameter will be automatically set to prevent spurious
        interactions between periodic images based on the cutoff,
        if the parameter is not already supplied by the user.

        To use partial atomic charges on the framework, set the `Atoms.set_initial_charges`
        method on the `Atoms` object. The charges will be written out to the CIF file
        and `UseChargesFromCIFFile` will be set to `yes` in the `simulation.input` file.

        To use no framework, have the `Atoms` object be empty, i.e. `Atoms()`.

        Parameters
        ----------
        profile
            An instantiated [raspa_ase.calculator.RaspaProfile][] object to use.
            The default is typically fine, which runs the following:
            `$RASPA_DIR/bin/simulate simulation.input`.
        directory
            The path to the directory to run the RASPA calculation in, which
            defaults to the current working directory.
        boxes
            A list of dictionaries, where each dictionary is a RASPA box.
            The default is an empty list.

            Example:

            ```python
            boxes = [{"BoxLengths": [25, 25, 25], "ExternalTemperature": 300.0, "Movies": True, "WriteMoviesEvery": 10}, {"BoxLengths": [30, 30, 30], "BoxAngles": [90, 120, 90], "ExternalTemperature": 500.0, "Movies": True, "WriteMoviesEvery": 10}]
            Raspa(boxes=boxes)
            ```

            would be written out as the following from 4.2 Example 2 of the RASPA manual:

            ```
            Box 0
                BoxLengths 25 25 25
                ExternalTemperature 300.0
                Movies yes
                WriteMoviesEvery 10
            Box 1
                BoxLengths 30 30 30
                BoxAngles 90 120 90
                ExternalTemperature 500.0
                Movies yes
                WriteMoviesEvery 10
            ```
        components
            A list of dictionaries, where each dictionary is a RASPA component.
            The default is an empty list.

            Example:

            ```python
            components = [{"MolelculeName": "N2", "MoleculeDefinition": "ExampleDefinition", "TranslationProbability": 1.0, "RotationProbability": 1.0, "ReinsertionProbability": 1.0, "CreateNumberOfMolecules": [50, 25]}, {"MoleculeName": "CO2", "MoleculeDefinition": "ExampleDefinitions", "TranslationProbability": 1.0, "RotationProbability": 1.0, "ReinsertionProbability": 1.0, "CreateNumberOfMolecules": [25, 50]}]
            Raspa(components=components)
            ```

            would be written out as the following from 4.2 Example 2 of the RASPA manual:

            ```
            Component 0 MoleculeName N2
                MoleculeDefinition ExampleDefinitions
                TranslationProbability 1.0
                RotationProbability 1.0
                ReinsertionProbability 1.0
                CreateNumberOfMolecules 50 25
            Component 1 MoleculeName CO2
                MoleculeDefinition ExampleDefinitions
                TranslationProbability 1.0
                RotationProbability 1.0
                ReinsertionProbability 1.0
                CreateNumberOfMolecules 25 50
            ```
        parameters
            Any RASPA parameters beyond the Box and Component parameters, formatted as a dictionary.
            Booleans will be converted to "Yes" or "No" automatically, and lists will be converted to
            space-separated strings. The RASPA parameters are case-insensitive.

            Example:

            ```python
            parameters = {"SimulationType": "MonteCarlo", "NumberOfCycles": 10000, "NumberOfInitializationCycles": 1000, "PrintEvery": 100, "ForceField": "ExampleMoleculeForceField"}
            Raspa(parameters=parameters)
            ```

            would be written out as the following from 4.2 Example 2 of the RASPA manual:

            ```
            SimulationType MonteCarlo
            NumberOfCycles 10000
            NumberOfInitializationCycles 1000
            PrintEvery 100
            Forcefield ExampleMoleculeForceField
            ```
        multiple_frameworks
            If you need to use multiple frameworks, then special treatment is required. Instead of
            having the framework as the `Atoms` object, you need to supply a list of `Atoms` objects
            to the `multiple_frameworks` keyword argument. In this case, the `Atoms` object
            the calculator is applied to should be empty, i.e. `Atoms()`.

            Example:

            ```python
            atoms1 = read("my_framework1.cif")
            atoms2 = read("my_framework2.cif")
            atoms1.info = {"UnitCells": [1, 1, 1]}
            atoms2.info = {"HeliumVoidFraction": 0.25, "UnitCells": [4, 4, 4]}
            Raspa(multiple_frameworks=[atoms1, atoms2])
            ```

            would be written out as the following:

            ```
            Framework 0
                FrameworkName framework0
                UnitCells 1 1 1
            Framework 1
                FrameworkName framework1
                HeliumVoidFraction 0.25
                UnitCells 4 4 4
            ```
        **kwargs
            Any additional arguments to pass to the `GenericFileIO` calculator.

        Returns
        -------
        None
        """

        profile = profile or RaspaProfile()
        multiple_frameworks = multiple_frameworks or []
        boxes = boxes or []
        components = components or []
        parameters = parameters or {}

        parameters = merge_parameters(
            parameters, get_framework_params(multiple_frameworks)
        )
        for i, component in enumerate(components):
            molecule_name = pop_parameter(component, "MoleculeName")
            parameters = merge_parameters(
                parameters, {f"Component {i} MoleculeName {molecule_name}": component}
            )
        for i, box in enumerate(boxes):
            parameters = merge_parameters(parameters, {f"Box {i}": box})

        super().__init__(
            template=RaspaTemplate(frameworks=multiple_frameworks),
            profile=profile,
            directory=directory,
            parameters=parameters,
            **kwargs,
        )



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def get_parameter(d: dict[str, Any], key: str, default: Any = None) -> Any:
    """
    Get a parameter from a dictionary, ignoring case.

    Parameters
    ----------
    d
        The dictionary to get the parameter from.
    key
        The parameter to get.
    default
        The default value to return if the parameter is not found.

    Returns
    -------
    Any
        The value of the parameter, or the default value if the parameter is not found.
    """
    d_lower = {k.lower(): v for k, v in d.items()}

    return d_lower.get(key.lower(), default)


def pop_parameter(d: dict[str, Any], key: str) -> Any:
    """
    Pop a parameter from a dictionary, ignoring case.

    Parameters
    ----------
    d
        The dictionary to pop the parameter from.
    key
        The parameter to pop.

    Returns
    -------
    Any
        The value of the parameter or None if the parameter is not found.
    """
    for k, v in d.items():
        if k.lower() == key.lower():
            del d[k]
            return v
    return None


def merge_parameters(dict1: dict[str, Any], dict2: dict[str, Any]) -> dict[str, Any]:
    """
    Merge two dictionaries, ignoring case.

    Parameters
    ----------
    dict1
        The first dictionary to merge.
    dict2
        The second dictionary to merge.

    Returns
    -------
    dict
        The merged dictionary.
    """
    merged_dict = deepcopy(dict1)

    for key2, value2 in dict2.items():
        matching_key = next(
            (key1 for key1 in merged_dict if key1.lower() == key2.lower()), None
        )

        if matching_key:
            merged_dict[matching_key] = value2
        else:
            merged_dict[key2] = value2

    return merged_dict


def convert_booleans(d: dict[str, Any]) -> dict[str, Any]:
    """
    Recursively convert boolean values to "yes" or "no".

    Parameters
    ----------
    d
        The dictionary to convert.

    Returns
    -------
    dict
        The dictionary with the boolean values converted to "Yes" or "No".
    """
    for key, value in d.items():
        if isinstance(value, bool):
            d[key] = "Yes" if value else "No"
        elif isinstance(value, dict):
            convert_booleans(value)
    return d



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from ase.atoms import Atoms

from raspa_ase.utils.dicts import get_parameter, merge_parameters

if TYPE_CHECKING:
    from typing import Any


def get_framework_params(frameworks: list[Atoms]) -> dict[str, Any]:
    """
    Get the framework-related parameters.

    Parameters
    ----------
    frameworks
        The frameworks to get the parameters for.

    Returns
    -------
    dict
        The framework-related parameters.
    """

    parameters = {}
    for i, framework in enumerate(frameworks):
        if framework == Atoms():
            continue

        name = f"framework{i}"
        cutoff = get_parameter(parameters, "CutOff", default=12.0)
        n_cells = get_suggested_cells(framework, cutoff)

        framework_params = {
            "FrameworkName": name,
            "UnitCells": n_cells,
        }

        if framework.has("initial_charges"):
            framework_params = merge_parameters(
                framework_params, {"UseChargesFromCIFFile": True}
            )

        parameters[f"Framework {i}"] = merge_parameters(
            framework_params,
            framework.info,
        )

    return parameters


def get_suggested_cells(framework: Atoms, cutoff: float) -> list[int]:
    """
    Get the suggested number of unit cells in each dimension for a given cutoff.

    Parameters
    ----------
    framework
        The framework to calculate the suggested number of unit cells for.
    cutoff
        The cutoff used for the calculation, in A.

    Returns
    -------
    tuple[int, int, int]
        The suggested number of unit cells in each dimension.
    """
    A, B, C = framework.get_cell()[:3]

    def _calculate_min_dist(v1, v2, v3):
        cross_product = np.cross(v1, v2)
        numerator = np.linalg.norm(np.dot(cross_product, v3))
        denominator = np.linalg.norm(cross_product)
        return np.divide(numerator, denominator)

    min_A = _calculate_min_dist(B, C, A)
    min_B = _calculate_min_dist(C, A, B)
    min_C = _calculate_min_dist(A, B, C)
    return [int(np.ceil(cutoff / (0.5 * min_i))) for min_i in [min_A, min_B, min_C]]



import os
from pathlib import Path

import pytest
from ase import Atoms
from ase.build import bulk
from ase.io import read

from raspa_ase.calculator import Raspa, RaspaProfile, RaspaTemplate

DATA_DIR = Path(__file__).parent / "data"


def test_profile_bad(monkeypatch):
    monkeypatch.delenv("RASPA_DIR", "/tmp")
    with pytest.raises(OSError):
        RaspaProfile()


def test_profile():
    profile = RaspaProfile()
    assert profile.binary == f"{os.getenv('RASPA_DIR')}/bin/simulate"

    assert profile.get_calculator_command() == [
        f"{os.getenv('RASPA_DIR')}/bin/simulate",
        "simulation.input",
    ]

    with pytest.raises(NotImplementedError):
        profile.version()


def test_profil2e():
    profile = RaspaProfile(binary="my/path")
    assert profile.binary == "my/path"

    assert profile.get_calculator_command() == [
        "my/path",
        "simulation.input",
    ]


def test_template():
    template = RaspaTemplate()
    assert template.inputname == "simulation.input"
    assert template.outputname == "raspa.stdout"
    assert template.errorfile == "raspa.stderr"


def test_notimplemented():
    template = RaspaTemplate()
    with pytest.raises(NotImplementedError):
        template.load_profile(None)


def test_template_execute(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    with Path(tmp_path / "simulation.input").open(mode="w") as fd:
        fd.write("")
    template = RaspaTemplate()
    template.execute(tmp_path, RaspaProfile())


def test_template_write_input(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    template = RaspaTemplate()
    template.write_input(
        RaspaProfile(), tmp_path, bulk("Cu"), {"CutOff": 12.8}, "energy"
    )
    assert (tmp_path / "simulation.input").exists()
    assert (
        tmp_path / "simulation.input"
    ).read_text() == "CutOff 12.8\nFramework 0\n    FrameworkName framework0\n    UnitCells 12 12 12\n"
    assert (tmp_path / "framework0.cif").exists()


def test_template_read_results():
    template = RaspaTemplate()
    assert template.read_results(".") == {"energy": None}


def test_raspa_blank():
    calc = Raspa()
    assert calc.directory == Path(".")
    assert calc.parameters == {}


def test_raspa_parameters():
    calc = Raspa(parameters={"CutOff": 12.8})
    assert calc.parameters == {"CutOff": 12.8}


def test_raspa_boxes():
    calc = Raspa(boxes=[{"BoxLengths": [1, 2, 3]}, {"BoxLengths": [4, 5, 6]}])
    assert calc.parameters == {
        "Box 0": {"BoxLengths": [1, 2, 3]},
        "Box 1": {"BoxLengths": [4, 5, 6]},
    }


def test_raspa_components():
    calc = Raspa(
        components=[
            {"MoleculeName": "N2", "MoleculeDefinition": "ExampleDefinition"},
            {
                "MoleculeName": "CO2",
                "MoleculeDefinition": "ExampleDefinition",
                "TranslationProbability": 1.0,
            },
        ]
    )
    assert calc.parameters == {
        "Component 0 MoleculeName N2": {"MoleculeDefinition": "ExampleDefinition"},
        "Component 1 MoleculeName CO2": {
            "MoleculeDefinition": "ExampleDefinition",
            "TranslationProbability": 1.0,
        },
    }


def test_raspa_functional1(tmp_path):
    atoms = Atoms()
    atoms.calc = Raspa(directory=tmp_path)
    atoms.get_potential_energy()
    assert Path(tmp_path / "simulation.input").exists()
    assert Path(tmp_path / "simulation.input").read_text() == ""


def test_raspa_functional2(tmp_path):
    atoms = bulk("Cu")
    atoms.set_initial_charges([1.0] * len(atoms))
    atoms.calc = Raspa(
        directory=tmp_path,
        boxes=[{"BoxLengths": [1, 2, 3]}, {"BoxLengths": [4, 5, 6]}],
        parameters={"CutOff": 12.8},
        components=[
            {"MoleculeName": "N2", "MoleculeDefinition": "ExampleDefinition"},
            {
                "MoleculeName": "CO2",
                "MoleculeDefinition": "ExampleDefinition",
                "TranslationProbability": 1.0,
            },
        ],
    )
    atoms.get_potential_energy()
    assert Path(tmp_path, "simulation.input").exists()
    assert (
        Path(tmp_path / "simulation.input").read_text()
        == "CutOff 12.8\nComponent 0 MoleculeName N2\n    MoleculeDefinition ExampleDefinition\nComponent 1 MoleculeName CO2\n    MoleculeDefinition ExampleDefinition\n    TranslationProbability 1.0\nBox 0\n    BoxLengths 1 2 3\nBox 1\n    BoxLengths 4 5 6\nFramework 0\n    FrameworkName framework0\n    UnitCells 12 12 12\n    UseChargesFromCIFFile Yes\n"
    )


def test_multi_frameworks(tmp_path):
    atoms1 = bulk("Cu")
    atoms1.info = {"HeliumVoidFraction": 0.75}
    atoms2 = bulk("Fe")
    atoms = Atoms()
    atoms.calc = Raspa(
        directory=tmp_path,
        multiple_frameworks=[atoms1, atoms2],
        boxes=[{"BoxLengths": [1, 2, 3]}, {"BoxLengths": [4, 5, 6]}],
        parameters={"CutOff": 12.8},
        components=[
            {"MoleculeName": "N2", "MoleculeDefinition": "ExampleDefinition"},
            {
                "MoleculeName": "CO2",
                "MoleculeDefinition": "ExampleDefinition",
                "TranslationProbability": 1.0,
            },
        ],
    )
    atoms.get_potential_energy()
    assert Path(tmp_path, "simulation.input").exists()
    input_str = Path(tmp_path / "simulation.input").read_text()
    assert (
        "Framework 0\n    FrameworkName framework0\n    UnitCells 12 12 12\n    HeliumVoidFraction 0.75\n"
        in input_str
    )
    assert (
        "Framework 1\n    FrameworkName framework1\n    UnitCells 12 12 12\n"
        in input_str
    )
    assert Path(tmp_path / "framework0.cif").exists()
    assert Path(tmp_path / "framework1.cif").exists()
    assert read(tmp_path / "framework0.cif")[0].symbol == "Cu"
    assert read(tmp_path / "framework1.cif")[0].symbol == "Fe"


@pytest.mark.skipif("RASPA_DIR" not in os.environ, reason="This test requires RASPA")
def test_example(tmp_path):
    atoms = Atoms()
    boxes = [
        {
            "BoxLengths": [30, 30, 30],
            "ExternalTemperature": 300,
            "Movies": True,
            "WriteMoviesEvery": 100,
        }
    ]
    components = [
        {
            "MoleculeName": "methane",
            "MoleculeDefinition": "ExampleDefinitions",
            "TranslationProbability": 1.0,
            "CreateNumberOfMolecules": 100,
        }
    ]
    parameters = {
        "SimulationType": "MonteCarlo",
        "NumberOfCycles": 100,
        "NumberOfInitializationCycles": 10,
        "PrintEvery": 10,
        "Forcefield": "ExampleMoleculeForceField",
    }
    calc = Raspa(
        directory=tmp_path, boxes=boxes, components=components, parameters=parameters
    )

    atoms.calc = calc
    atoms.get_potential_energy()
    assert "System_0" in calc.results
    assert "output_Box_1.1.1_300.000000_0.data" in calc.results["System_0"]
    assert calc.results["System_0"]["output_Box_1.1.1_300.000000_0.data"]["Simulation"][
        "Dimensions"
    ] == [3.0]


@pytest.mark.skipif("RASPA_DIR" not in os.environ, reason="This test requires RASPA")
def test_example2(tmp_path):
    atoms = Atoms()
    boxes = [
        {
            "BoxLengths": [25, 25, 25],
            "ExternalTemperature": 300.0,
            "Movies": True,
            "WriteMoviesEvery": 10,
        },
        {
            "BoxLengths": [30, 30, 30],
            "BoxAngles": [90, 120, 120],
            "ExternalTemperature": 500,
            "Movies": True,
            "WriteMoviesEvery": 10,
        },
    ]
    components = [
        {
            "MoleculeName": "N2",
            "MoleculeDefinition": "ExampleDefinitions",
            "TranslationProbability": 1.0,
            "RotationProbability": 1.0,
            "ReinsertionProbability": 1.0,
            "CreateNumberOfMolecules": [50, 25],
        },
        {
            "MoleculeName": "CO2",
            "MoleculeDefinition": "ExampleDefinitions",
            "TranslationProbability": 1.0,
            "RotationProbability": 1.0,
            "ReinsertionProbability": 1.0,
            "CreateNumberOfMolecules": [25, 50],
        },
    ]
    parameters = {
        "SimulationType": "MonteCarlo",
        "NumberOfCycles": 20,
        "NumberOfInitializationCycles": 10,
        "PrintEvery": 10,
        "Forcefield": "ExampleMoleculeForceField",
    }
    calc = Raspa(
        directory=tmp_path, boxes=boxes, components=components, parameters=parameters
    )

    atoms.calc = calc
    atoms.get_potential_energy()
    assert "System_0" in calc.results
    assert "System_1" in calc.results
    assert "output_Box_1.1.1_300.000000_0.data" in calc.results["System_0"]
    assert "output_Box_1.1.1_500.000000_0.data" in calc.results["System_1"]


@pytest.mark.skipif("RASPA_DIR" not in os.environ, reason="This test requires RASPA")
def test_example3(tmp_path):
    atoms = read(Path(DATA_DIR / "MFI_SI.cif"))
    atoms.info = {
        "UnitCells": [2, 2, 2],
        "HeliumVoidFraction": 0.29,
        "ExternalTemperature": 300.0,
        "ExternalPressure": [1e4, 1e5],
    }
    components = [
        {
            "MoleculeName": "methane",
            "MoleculeDefinition": "ExampleDefinitions",
            "TranslationProbability": 0.5,
            "ReinsertionProbability": 0.5,
            "SwapProbability": 1.0,
            "CreateNumberOfMolecules": 0,
        }
    ]
    parameters = {
        "SimulationType": "MonteCarlo",
        "NumberOfCycles": 25,
        "NumberOfInitializationCycles": 2,
        "PrintEvery": 5,
        "Forcefield": "ExampleZeolitesForceField",
        "RemoveAtomNumberCodeFromLabel": True,
        "ComputeNumberOfMoleculesHistogram": True,
        "WriteNumberOfMoleculesHistogramEvery": 5,
        "NumberOfMoleculesHistogramSize": 5,
        "NumberOfMoleculesRange": 5,
        "ComputeEnergyHistogram": True,
        "WriteEnergyHistogramEvery": 5,
        "EnergyHistogramSize": 5,
        "EnergyHistogramLowerLimit": -110000,
        "EnergyHistogramUpperLimit": -20000,
    }
    calc = Raspa(directory=tmp_path, components=components, parameters=parameters)

    atoms.calc = calc
    atoms.get_potential_energy()
    assert "System_0" in calc.results
    assert "output_framework0_2.2.2_300.000000_10000.data" in calc.results["System_0"]
    assert "output_framework0_2.2.2_300.000000_100000.data" in calc.results["System_0"]
